/************************************************************************************
 * SpriteRender.c
 *
 * CheeseToast by Jim Bumgardner
 *
 * These are the routines used to speed sprite rendering.
 *
 * 2/28/95
 * Once upon a time they were in 68000 assembly language,
 * but I have converted them to C, since they seem to run
 * fast enough this way, and will compile for the Power PC.
 *
 ************************************************************************************/

#include "CToast.h"
#include <OSUtils.h>

#define ERASEBLACK	1

extern SpriteDef sDef[];
extern SpriteInstance *sTable;
extern long		gScreenRowBytes,gVideoRowBytes;
extern Ptr		gScreenMem,gVideoMem;
extern Rect		gPlayRect,gOffscreenRect;

void LSpriteDraw(register SpriteInstance *sp)
{
	unsigned char	*sprPtr,*scrPtr;
	LSprite			*lSpriteHeader;
	// long			*maskPtr;
	short			j,k;
	unsigned char	typ,cnt;
	long			rowBytesOffset;
	Point			loc;
	
	loc = sp->pos;
	loc.h += (gUnivWidth - gOffsetPt.h); loc.h %= gUnivWidth;
	loc.v += (gUnivHeight - gOffsetPt.v); loc.v %= gUnivHeight;
	if (!PtInRect(loc,&gPlayRect))
		return;

	sp->update = true;

	lSpriteHeader = (LSprite *) sDef[sp->type].colorMaps;

	rowBytesOffset = gScreenRowBytes - lSpriteHeader->width;	// assumed to be multiple of 4

	// Copy area underneath sprite to temporary save area (part of sprite structure)
	// so sprite can be erased.
	// Pixels are copied 4 at a time, loop is unrolled for speed
	//


	sprPtr = (unsigned char *) &lSpriteHeader->data[lSpriteHeader->frameOffset[sp->aniState]];
	scrPtr = (unsigned char *) (gScreenMem + (loc.v*gScreenRowBytes + loc.h));

	for (j = lSpriteHeader->height; j; j--) {
		k = lSpriteHeader->width;
		sprPtr += sizeof(short);	// Skip Count
		while (k) {
			typ = *sprPtr & LSPR_HiBits;
			cnt = *sprPtr & LSPR_LowBits;
			k -= cnt;
			++sprPtr;
			switch (typ) {
			case LSPR_Mask:
				scrPtr += cnt;
				break;
			case LSPR_Opaque:
				while (cnt--)
					*(scrPtr++) = *(sprPtr++);
				break;
			case LSPR_Mixed: // unsupported
				while (cnt--)
					*(scrPtr++) = *(sprPtr++);
				break;
			}
		}
		scrPtr += rowBytesOffset;
	}
}


void LargeSpriteErase(register SpriteInstance *sp)
{
	long	*scrnPtr;
	long 	rowBytesOffset;
	Point	loc;
	short	n;

	loc = sp->pos;
	loc.h += (gUnivWidth - gOldOffset.h); loc.h %= gUnivWidth;
	loc.v += (gUnivHeight - gOldOffset.v); loc.v %= gUnivHeight;
	if (!PtInRect(loc,&gPlayRect))
		return;

	sp->update = true;

	scrnPtr = (long *) (gScreenMem + (loc.v*gScreenRowBytes + loc.h));
	rowBytesOffset = gScreenRowBytes - LargeIconWidth;
	rowBytesOffset >>= 2;


	n = LargeIconHeight;
	while (n--) {
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		scrnPtr += rowBytesOffset;
	}
}

void MedSpriteErase(register SpriteInstance *sp)
{
	long	*scrnPtr;
	long 	rowBytesOffset;
	Point	loc;
	short	n;

	loc = sp->pos;
	loc.h += (gUnivWidth - gOldOffset.h); loc.h %= gUnivWidth;
	loc.v += (gUnivHeight - gOldOffset.v); loc.v %= gUnivHeight;
	if (!PtInRect(loc,&gPlayRect))
		return;

	sp->update = true;


	scrnPtr = (long *) (gScreenMem + (loc.v*gScreenRowBytes + loc.h));
	rowBytesOffset = gScreenRowBytes - IconWidth;
	rowBytesOffset >>= 2;


	n = IconHeight;
	while (n--) {
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		scrnPtr += rowBytesOffset;
	}
}

void CreamBulletErase(register SpriteInstance *sp)
{
	long	*scrnPtr;
	long 	rowBytesOffset;
	Point	loc;
	short	n;

	loc = sp->pos;
	loc.h += (gUnivWidth - gOldOffset.h); loc.h %= gUnivWidth;
	loc.v += (gUnivHeight - gOldOffset.v); loc.v %= gUnivHeight;
	if (!PtInRect(loc,&gPlayRect))
		return;

	sp->update = true;


	scrnPtr = (long *) (gScreenMem + (loc.v*gScreenRowBytes + loc.h));
	rowBytesOffset = gScreenRowBytes - 16;
	rowBytesOffset >>= 2;


	n = 16;
	while (n--) {
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		*(scrnPtr++) = 0xFFFFFFFF;
		scrnPtr += rowBytesOffset;
	}
}


// Draw a Bullet
//

unsigned short bulletColor[35] = {0xfefe,0xfdfd,0xfcfc,0xfcfc,0xfbfb,
								  0xfbfb,0xfafa,0xfafa,0xf9f9,0xf9f9,
								  0xf8f8,0xf8f8,0xf8f8,0xf7f7,0xf7f7,
								  0xf7f7,0xf6f6,0xf6f6,0xf6f6,0xf5f5,
								  0xf5f5,0xf5f5,0x0000,0x0000,0x0000,
								  0x0000,0x0000,0x0000,0x0000,0x0000,
								  0x0000,0x0000,0x0000,0x0000,0x0000};

void BulletDraw(register SpriteInstance *sp)
{
	short	*scrnPtr;
	long	rowBytesOffset;
	unsigned short clr;
	Point	loc;

	clr = bulletColor[sp->lifeSpan];
	loc = sp->pos;
	loc.h += (gUnivWidth - gOffsetPt.h); loc.h %= gUnivWidth;
	loc.v += (gUnivHeight - gOffsetPt.v); loc.v %= gUnivHeight;
	if (!PtInRect(loc,&gPlayRect))
		return;

	sp->update = true;

	rowBytesOffset = gScreenRowBytes - BulletWidth;
	rowBytesOffset >>= 1;

	scrnPtr = (short *) (gScreenMem + (loc.v*gScreenRowBytes + loc.h));

	// Draw Bullet, 2 pixels at a time (pixels are cleared to 0 = white)
	*(scrnPtr++) = clr;
	scrnPtr += rowBytesOffset;
	*scrnPtr = clr;
}

// Draw a Yellow Bullet
//

void PhotonDraw(register SpriteInstance *sp)
{
	short	*scrnPtr;
	long	rowBytesOffset;
	Point	loc;

	loc = sp->pos;
	loc.h += (gUnivWidth - gOffsetPt.h); loc.h %= gUnivWidth;
	loc.v += (gUnivHeight - gOffsetPt.v); loc.v %= gUnivHeight;
	if (!PtInRect(loc,&gPlayRect))
		return;

	sp->update = true;

	rowBytesOffset = gScreenRowBytes - BulletWidth;
	rowBytesOffset >>= 1;
	
	scrnPtr = (short *) (gScreenMem + (loc.v*gScreenRowBytes + loc.h));

	// Draw Photons 2 pixels at a time (05 = Yellow)
	*(scrnPtr++) = 0x0505;
	scrnPtr += rowBytesOffset;
	*scrnPtr = 0x0505;
}

// Erase a bullet, by restoring the area that was saved
//
void BulletErase(register SpriteInstance *sp)
{
	short	*scrnPtr;
	long	rowBytesOffset;

	Point	loc;

	loc = sp->pos;
	loc.h += (gUnivWidth - gOldOffset.h); loc.h %= gUnivWidth;
	loc.v += (gUnivHeight - gOldOffset.v); loc.v %= gUnivHeight;
	if (!PtInRect(loc,&gPlayRect))
		return;

	sp->update = true;

	scrnPtr = (short *) ((char *) gScreenMem + (loc.v*gScreenRowBytes + loc.h));
	rowBytesOffset = gScreenRowBytes - BulletWidth;
	rowBytesOffset >>= 1;

	*(scrnPtr++) = 0xFFFF;
	scrnPtr += rowBytesOffset;
	*scrnPtr = 0xFFFF;
}

// Draw a spark (single pixel)
//
void SparkDraw(register SpriteInstance *sp)
{
	register Ptr	scrnPtr;
	register char	color;
	register Point	loc;

	loc = sp->pos;
	loc.h += (gUnivWidth - gOffsetPt.h); loc.h %= gUnivWidth;
	loc.v += (gUnivHeight - gOffsetPt.v); loc.v %= gUnivHeight;
	if (!PtInRect(loc,&gPlayRect))
		return;

	sp->update = true;

	color = sp->state;
	scrnPtr = gScreenMem + (loc.v*gScreenRowBytes + loc.h);
	*scrnPtr = color;				// Draw spark
}

// Erase a spark
//
void SparkErase(register SpriteInstance *sp)
{
	register Ptr	scrnPtr;
	register Point	loc;

	loc = sp->pos;
	loc.h += (gUnivWidth - gOldOffset.h); loc.h %= gUnivWidth;
	loc.v += (gUnivHeight - gOldOffset.v); loc.v %= gUnivHeight;
	if (!PtInRect(loc,&gPlayRect))
		return;

	sp->update = true;

	scrnPtr = gScreenMem + (loc.v*gScreenRowBytes + loc.h);
	*scrnPtr = 0xFF;
}


// Copy entire offscreen area to onscreen window
//
void MyCopyBits(void)
{
	register long	*scrnPtr,*vidPtr;
	register long	vidRowBytesOffset,scrnRowBytesOffset;
	register short	y;
#if THINK_C
	char			mmode;
#else
	SInt8			mmode;
#endif

	if (g12InchMode) {
		// If using a 12 inch monitor (512 x 384), use special version of
		// this routine
		MyCopyBits12Inch();
		return;
	}
	// otherwise, assume 640 x 480

	scrnPtr = (long *) ((char *) gScreenMem + (SafeWidth + (gScreenRowBytes * SafeHeight)));
	scrnRowBytesOffset = gScreenRowBytes - 640;
	scrnRowBytesOffset >>= 2;

	vidPtr = (long *) (gVideoMem + (gVidOffset.v * gVideoRowBytes) + gVidOffset.h);
	vidRowBytesOffset = gVideoRowBytes - 640;
	vidRowBytesOffset >>= 2;

	// When drawing to onscreen video, we must temporarily switch to
	// 32bit memory, since video address will generally be > 8 megs
	//
	mmode = true32b;
	SwapMMUMode(&mmode);
	y = 480;
	while (y--) {
		// 160 occurances
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		
		vidPtr += vidRowBytesOffset;
		scrnPtr += scrnRowBytesOffset;
	}
	SwapMMUMode(&mmode);
}

void MyCopyBits12Inch(void)
{
	register long	*scrnPtr, *vidPtr;
	register long	vidRowBytesOffset,scrnRowBytesOffset;
	register short	y;

	SInt8			mmode;

	scrnPtr = (long *) (gScreenMem + (SafeWidth + (gScreenRowBytes * SafeHeight)));
	scrnRowBytesOffset = gScreenRowBytes - 512;
	scrnRowBytesOffset >>= 2;

	vidPtr = (long *) (gVideoMem + (gVidOffset.v * gVideoRowBytes) + gVidOffset.h);
	vidRowBytesOffset = gVideoRowBytes - 512;
	vidRowBytesOffset >>= 2;

	// When drawing to onscreen video, we must temporarily switch to
	// 32bit memory, since video address will generally be > 8 megs
	//
	mmode = true32b;
	SwapMMUMode(&mmode);
	y = 384;
	while (y--) {
		// 128 occurances
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);

		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		*(vidPtr++) = *(scrnPtr++);
		vidPtr += vidRowBytesOffset;
		scrnPtr += scrnRowBytesOffset;
	}
	SwapMMUMode(&mmode);
}

// This routine is used to copy a rectangle (typically a sprite)
// from offscreen pixelmap to onscreen video
//
void MyCopyRect(register Rect *r)
{
	register Ptr	scrnPtr, vidPtr;
	register long	vidRowBytesOffset,scrnRowBytesOffset;
	register short	y,x;
	SInt8			mmode;

	scrnPtr = gScreenMem;
	scrnPtr += r->left + (gScreenRowBytes *r->top);
	scrnRowBytesOffset = gScreenRowBytes - (r->right - r->left);

	vidPtr = gVideoMem + (gVidOffset.v * gVideoRowBytes) + gVidOffset.h;
	vidPtr += (r->left-SafeWidth) + ((r->top-SafeHeight)*gVideoRowBytes);
	vidRowBytesOffset = gVideoRowBytes - (r->right - r->left);

	// When drawing to onscreen video, we must temporarily switch to
	// 32bit memory, since video address will generally be > 8 megs
	//
	mmode = true32b;
	SwapMMUMode(&mmode);
	y = r->bottom - r->top;
	while (y--) {
		x = r->right - r->left;
		while (x--) {
			*(vidPtr++) = *(scrnPtr++);
		}
		scrnPtr += scrnRowBytesOffset;
		vidPtr += vidRowBytesOffset;
	}
	SwapMMUMode(&mmode);
}
