#include "CToast.h"
#include <Events.h>
#include "ToastTalk.h"

/*************************************************************************
 * Cmds (Server -> Clients)
 *
 *		sLis	Sprite list
 *		iGam	Init Game (Begin Game Mode)
 *		eGam	End Game (End Game)
 *		nSpr	New Sprite
 *		kSpr	Kill Sprite
 *		eSpr	Explode Sprite
 *		sSpr	Sprite State Change
 *		rSem	Request Semaphore (for frame to frame synchronization)
 *
 *
 * Client -> Server
 *		regi	Sign On (Client to Server)
 *!!	bye		Sign Off (Unimplemented!!)
 *		kybd	Keyboard Activity
 *		aSem	Acknoledge Semaphore
 *


	Bugs:
		Client can move server's ship, but client can't see server's ship.
		Client's star movement doesn't scroll when server ship is moving.

		When Client Quits (trace through step by step and see why the following happens)
			Client goes back to main loop, but screen graphics don't change
			   (client may not be executing game over code...)

			Server locks up in ServerIdle.

 *************************************************************************/

// Problems
//
// Current version uses old records in sprite table to determine erase squares
//  this causes potential refresh / efficency problems with current model of having 
//  new sprite lists sent by server
//



SendSpriteTable	*sendTab;

void ReceiveSpriteList(SendSpriteTable *sTab);
void ReceiveKeys(long clientID, short keyFlags);

Boolean	gClientLock,gServerLock;

Boolean	gSessionID;	// Used by Client

void ProcessToastEvent(long cmd, long msgRefCon, char *buffer, long len)
{
	switch (cmd) {

//
// Server -> Client Events
//
	case 'sLis':	// Sprite list (server->client)
		ReceiveSpriteList((SendSpriteTable *) buffer);
		break;
	case 'ackr':	// Register Client
		gSessionID = msgRefCon;
		break;
	case 'iGam':
		gClientMode = true;
		BeginGame();
		break;
		break;
	case 'eGam':
		EndGame();
		break;
	case 'nSpr':	// New Sprite (ID# Sprite Struc) (Server->Client)
		ReceiveNewSprite((SpriteInstance *) buffer);
		break;
	case 'kSpr':	// Kill Sprite (ID#) (Server->Client)
		ReceiveKillSprite(*((long *) buffer));
		break;
	case 'eSpr':	// Explode Sprite (ID# - Explosion) (Server->Client)
		ReceiveExplodeSprite(*((long *) buffer), *((short *) &buffer[4]));
		break;
	case 'sSpr':	// Sprite State Change (ID#) (SpriteState) (Server->Client)
		ReceiveSpriteStateChange((SpriteInstance *) buffer);
		break;
	case 'rSem':	// Request Semaphore (Server->Client)
		gClientLock = false;
		ClientToServerEvent('aSem',gSessionID,NULL,0L);
		break;
//
// Client->Server Events
//
	case 'kybd':	// Keyboard State (Kybd Flags) (Client -> Server)
		ReceiveKeys(msgRefCon, *((short *) buffer));
		break;
	case 'aSem':	// Acknowledge Semaphore (Client->Server)
		gServerLock = false;
		break;
	case 'quit':	// Kill that client...
		DeactivateClient(msgRefCon);
		break;
	default:
		break;
	}
}

void ReceiveKeys(long clientID, short keyFlags)
{
	// Client Keys Received - use to modify client ships...
	// !!! for now, client gets to modify gKeyFlags...
	gKeyFlags |= keyFlags;
}

void ReceiveSpriteList(SendSpriteTable *sTab)
{
	short			i;
	SpriteInstance	*sp;
	for (i = 0,sp = sTable; i < gMaxSprite; ++i,++sp)
		KillSprite(sp);

	gMaxSprite = 0;
	gSpriteCnt = 0;
	for (i = 0,sp=sTable; i < sTab->nbrSendSprites; ++i,++sp) {
		*sp = sTab->sendTable[i];
		if (sp->type >= ST_Teapot && sp->type <= ST_TeapotNew)
			gShip = sp;
	}
	gSpriteCnt = gMaxSprite = sTab->nbrSendSprites;
}

void ReceiveSpriteStateChange(SpriteInstance *si)
{
	SpriteInstance	*sp;
	short			i;
	for (i = 0,sp=sTable; i < gMaxSprite; ++i,++sp) {
		if (sp->id == si->id) {
			if (sp == gShip) {
				if (sp->type== ST_TeapotNew && si->type == ST_Teapot)
					PlaySound(S_ShipDebut, 3);
				if (si->type == ST_TeapotT && sp->type != ST_TeapotT)
					PlaySound(S_Thrust, 1);
				if (si->type == ST_TeapotS && sp->type != ST_TeapotS)
					PlaySound(S_Shield, 1);
				if (sp == gShip && gShip->state & SF_Fire)
					PlaySound(S_Firing, 1);
			}
			*sp = *si;
			return;
		}
	}
}

void ReceiveNewSprite(SpriteInstance *si)
{
	SpriteInstance	*sp;
	// short			i;

	sp = &sTable[gMaxSprite];
	*sp = *si;

	// !!! Only for your OWN ship...
	if (sp->type >= ST_Teapot && sp->type <= ST_TeapotNew)
		gShip = sp;

	++gMaxSprite;
	++gSpriteCnt;
	switch (si->type) {
	case ST_Saucer:
	case ST_Beater:
	case ST_Mixer:
	case ST_Croissant:
	case ST_Cube:
		PlaySound(S_BadGuyDebut, 3);
		break;
	case ST_CoffeePot:
		PlaySound(S_KevinsPot, 3);
		break;
	case ST_Photon:
		PlaySound(S_EnemyFires, 2);
		break;
	}
}

void ReceiveExplodeSprite(long id, short sparkLifeSpan)
{
	short			i;
	SpriteInstance	*sp;
	for (i = 0,sp=sTable; i < gMaxSprite; ++i,++sp) {
		if (sp->id == id) {
			ExplodeSprite(sp,sparkLifeSpan);
			return;
		}
	}
}

void ReceiveKillSprite(long id)
{
	short			i;
	SpriteInstance	*sp;
	for (i = 0,sp=sTable; i < gMaxSprite; ++i,++sp) {
		if (sp->id == id) {
			// or Creamer
			switch (sp->type) {
			case ST_Yummies:
			case ST_Creamer:
				PlaySound(S_YummyConsume, 3);
				break;
			case ST_Wheel:
				PlaySound(S_BreakRings, 2);
				break;
			case ST_Aster:
				PlaySound(S_CompoundFracture, 2);
				break;
			case ST_Teacup:
				PlaySound(S_BreakCups, 2);
				break;
			case ST_BigJim:
				PlaySound(S_JimOuch, 2);
				break;
			}
			KillSprite(sp);
			return;
		}
	}
}


void SendSpriteList(ClientPtr	cp)
{
	short			i;
	SpriteInstance	*sp;

	if (sendTab == NULL)
		sendTab = (SendSpriteTable *) NewPtrClear(sizeof(SendSpriteTable));
	sendTab->nbrSendSprites = 0;
	for (i = 0,sp = sTable; i < gMaxSprite; ++i,++sp) {
		if (sp->active)
			sendTab->sendTable[sendTab->nbrSendSprites++] = *sp;
	}
	ServerToClientEvent(cp, 'sLis',0L,(Ptr) sendTab, sizeof(short) + sendTab->nbrSendSprites * sizeof(SpriteInstance));
}

void SendSpriteStateChange(SpriteInstance *sp)
{
	if (gActiveClients && !gClientMode)
		ServerToAllEvent('sSpr',0L,(Ptr) sp, sizeof(SpriteInstance));
}

void SendNewSprite(SpriteInstance *sp)
{
	if (gActiveClients && !gClientMode)
		ServerToAllEvent('nSpr',0L,(Ptr) sp, sizeof(SpriteInstance));
}

void SendKillSprite(long id)
{
	if (gActiveClients && !gClientMode)
		ServerToAllEvent('kSpr',0L,(Ptr) &id, sizeof(long));
}

void SendExplodeSprite(long id, short sparkLifespan)
{
	char	buffer[6];
	
	if (gActiveClients && !gClientMode) {
		*((long *) &buffer[0]) = id;
		*((short *) &buffer[4]) = sparkLifespan;
		ServerToAllEvent('eSpr',0L,(Ptr) buffer, 6L);
	}
}


void InitGameEvent()
{
	// short		n;
	// ClientPtr	cp;

	if (gNbrClients && !gClientMode) {
		ActivateClients();
		ServerToAllEvent('iGam',0L,NULL,0L);
		// for (n = 0,cp = clients; n < gNbrClients; ++n,++cp)
			// SendSpriteList(cp);
	}
}

void EndGameEvent()
{
	if (gActiveClients && !gClientMode)
		ServerToAllEvent('eGam',0L,NULL,0L);
}

void ServerIdle()
{
	EventRecord	theEvent;
	if (!gClientMode && gConnectionType && gActiveClients) {
		// Send Request Semaphore...
		//
		ServerToAllEvent('rSem',0L,NULL,0L);

		// Wait for Clients to Respond
		gServerLock = true;
		while (gServerLock == true && gGameState == GS_Play && gActiveClients) {
			while (EventAvail(highLevelEventMask,&theEvent)) {
				WaitNextEvent(highLevelEventMask,&theEvent,0L,NULL);
				if (theEvent.what == kHighLevelEvent && theEvent.message == ToastEventID)
					ProcessAppleTalkEvent(&theEvent);
			}
		}
	}
}

void ComputeViewOffset()
{
	register SpriteInstance	*sp;
	sp = gShip;
	if (sp) {
		// Everything is erased - move scroll offset
		gOffsetPt.h += (sp->vector.lh >> 16);
		gOffsetPt.v += (sp->vector.lv >> 16);
		if (gOffsetPt.h < 0)					gOffsetPt.h += gUnivWidth;
		if (gOffsetPt.v < 0)					gOffsetPt.v += gUnivHeight;
		if (gOffsetPt.h >= gUnivWidth)			gOffsetPt.h -= gUnivWidth;
		if (gOffsetPt.v >= gUnivHeight)			gOffsetPt.v -= gUnivHeight;
	}
}

void ClientIdle()
{
	EventRecord	theEvent;
	short		i,n;
	register SpriteInstance	*sp;
	static		short	lastKybdFlags;

	if (gClientMode) {
		if (gGameState == GS_GameOver) {
			ClientToServerEvent('quit',gSessionID,NULL,0L);
			return;
		}
		// Move all sprites
		//
		n = gMaxSprite;
		for (i = 0,sp=sTable; i < n; ++i,++sp) {
			// Render Sprite in New Position
			StandardSpriteMove(sp);
		}

		// Check Keyboard state...
		if (lastKybdFlags != gKeyFlags) {
			lastKybdFlags = gKeyFlags;
			ClientToServerEvent('kybd',gSessionID,(Ptr) &lastKybdFlags,2L);
		}

		// Receive Event Updates (New Sprites and Such)
		//
		
		gClientLock = true;
		while (gClientLock == true && gGameState == GS_Play) {
			while (EventAvail(highLevelEventMask,&theEvent)) {
				WaitNextEvent(highLevelEventMask,&theEvent,0L,NULL);
				if (theEvent.what == kHighLevelEvent && theEvent.message == ToastEventID)
					ProcessAppleTalkEvent(&theEvent);
			}
		}

		ComputeViewOffset();
	}
}
