////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Asset.h - an implementation of something like Apples Resource Manager
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef _H_Asset
#define _H_Asset        1

#include "local.h"

#ifdef macintosh
  #define ASSET_CREATOR    'mPrp'
  #define ASSET_FILETYPE   'mAst'
  #define ASSET_FILETYPEPC 'tsAm'

  #define ASSET_FAVE       'Fave'
  #define RT_PROP          'Prop'
#else
  #define ASSET_CREATOR    0x00000000L    // 'mPrp'
  #define ASSET_FILETYPE   0x00000000L    // 'mAst'

  #define ASSET_FAVE       0x46617665L     // 'Fave'

  #define RT_PROP          0x50726f70L    // 'Prop'
#endif


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef short   ErrNbr;
typedef long    AssetType;

typedef struct {
        long    dataOffset;         // Swap
        long    dataSize;           // Swap
        long    assetMapOffset;     // Swap
        long    assetMapSize;       // Swap
} AssetFileHeader;

typedef struct {
        long    nbrTypes;
        long    nbrAssets;
        long    lenNames;
        long    typesOffset;    // Offset into AssetMap
        long    recsOffset;             // offset into AssetMap
        long    namesOffset;
} AssetMapHeader;

typedef struct {
        AssetType       assetType;
        long            nbrAssets;
        long            firstAsset;     // Index # in Asset Recs List
} AssetTypeRec;

typedef struct {
        long      idNbr;        // Unique ID#
        Handle    rHandle;		// (used for Handle)
        long      dataOffset;
        long      dataSize;
        long      lastUseTime;  // seconds since 1/1/04
        long      nameOffset;   // offset into StringList or -1
        long      flags;
        unsigned long  crc;		// This is a refnum 6/9/95
} AssetRec;

typedef struct AssetFileVars {
        struct AssetFileVars *nextAssetFile;
        short                   aRefNum;
        AssetFileHeader afHeader;
        AssetMapHeader  afMap;
        Handle                  typeList;
        Handle                  assetList;
        Handle                  nameList;
        FSSpec                  fsSpec;
        Boolean                 fileNeedsUpdate;
} AssetFileVars;

#define AssetModified   0x01
#define AssetLoaded             0x02
#define AssetPurgeable  0x04
#define AssetProtected  0x08

typedef struct {
	long	id;
	unsigned long crc;
} AssetSpec, *AssetSpecPtr;	// 6/7/95 JAB

typedef struct {
        AssetType       type;
		AssetSpec		spec;	// 6/7/95 JAB
        long            blockSize;
        long            blockOffset;
        short           blockNbr;
        short           nbrBlocks;
        union {
                struct {
                        long    flags;
                        long    size;
                        char    name[32];
                        char    data[1];
                } firstBlockRec;
                struct {
                        char    data[1];
                } nextBlockRec;
        } varBlock;
} AssetBlockHeader, *AssetBlockPtr;     // Used to send assets over net

// Shared Routines
//

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// ErrorHandling
ErrNbr  AssetError();   // returns last error

// Asset File Management
ErrNbr  OpenAssetFile(FSSpec *sfFile, short *refNum);
ErrNbr  NewAssetFile(FSSpec *sfFile, short *refNum);
ErrNbr  CloseAssetFile(short refNum);
ErrNbr  WriteAssetFile();
ErrNbr  UpdateAssetFile();
short   CurAssetFile();
ErrNbr  UseAssetFile(short refNum);
ErrNbr  UpdateAssetFile(void);

// Asset Maintainence
Handle  GetAsset(AssetType type, long id);
ErrNbr  ChangedAsset(Handle h);
ErrNbr  AddAsset(Handle h, AssetType type, long id, StringPtr name);
ErrNbr  GetAssetInfo(Handle h, AssetType *type, long *id, StringPtr name);
ErrNbr  SetAssetInfo(Handle h, long id, StringPtr name);

// Asset Memory Management
ErrNbr  DetachAsset(Handle h);
ErrNbr  ReleaseAsset(Handle h);

// Asset Queries
long    CountAssets(AssetType type);
Handle  GetIndAsset(AssetType type, long nbr);
long    CountAssetTypes();
AssetType GetIndAssetType(long nbr);
Boolean AssetExists(AssetType type, long id);
long    UniqueAssetID(AssetType type, long minID);

// Internal Asset Manager functions
AssetRec        *FindAssetRec(AssetType type, long id);
AssetRec        *FindAssetRecByIndex(AssetType type, long idx);
Boolean         FindAssetByHandle(Handle rHandle, AssetTypeRec **att, AssetRec **app);

// CRC Functions 6/7/95
unsigned long 	ComputeCRC(Ptr p, long len); // 6/7/95
AssetRec		*FindAssetRecWithCRC(AssetType type, long id, unsigned long crc);	// 6/7/95
Handle			GetAssetWithCRC(AssetType type, long id, unsigned long crc);	// 6/7/95
unsigned long	GetAssetCRC(Handle h);
Boolean 		AssetExistsWithCRC(AssetType type, long id, unsigned long crc);
ErrNbr 			SetAssetCRC(Handle h, unsigned long crc);

// Asset by Name functions 6/7/95
AssetRec	*FindAssetRecByName(AssetType type, StringPtr name);
Handle	GetAssetByName(AssetType type, StringPtr name);

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
ErrNbr NewAssetFileRec(short refNum);
ErrNbr LoadAssetMap(short refNum);
ErrNbr WriteAsset(AssetRec *ap);
ErrNbr WriteAssetMap(short oRefNum);
ErrNbr RmveAsset(Handle rHandle);


#endif

